//******************************************************************************
// FILE..........: BTreeArray.hpp
// DESCRIPTION...: B-tree based array
// AUTHOR........: Dmitry Vinogradov
// COPYRIGHT.....: Copyright(c) FastCpp 2012-2014, Dmitry Vinogradov
// LICENSE.......: "2 EUR" or "advertising note at the about box"
//                 See http://www.FastCpp.com/license.html
//
// HISTORY.......: DATE       COMMENT
//                 ---------- -------------------------------------------------
//                     y  m  d
//                 2012-11-17 Created by Dmitry Vinogradov
//******************************************************************************

//---------------------------------------------------------------- include guard
#ifndef FastCpp_BTREEARRAY_HPP
#define FastCpp_BTREEARRAY_HPP

//--------------------------------------------------------------- import std lib
#include <cstring>
#include <limits>
#include <stdexcept>
#include <utility>
#include <vector>

//------------------------------------------------------------------------------
#ifndef                     FastCpp_BTREEARRAYUPBLOCK_HPP
#include <FastCpp/BTreeArray/detail/BTreeArrayUpBlock.hpp>
#endif//                    FastCpp_BTREEARRAYUPBLOCK_HPP

#ifndef              FastCpp_BTREEARRAYITERATOR_HPP
#include <FastCpp/BTreeArray/BTreeArrayIterator.hpp>
#endif//             FastCpp_BTREEARRAYITERATOR_HPP

#ifndef              FastCpp_BTREEARRAYCONSTITERATOR_HPP
#include <FastCpp/BTreeArray/BTreeArrayConstIterator.hpp>
#endif//             FastCpp_BTREEARRAYCONSTITERATOR_HPP

#ifndef                 FASTCPP_ISINTEGRAL_HPP
#include <FastCpp/Metafunctions/IsIntegral.hpp>
#endif//                FASTCPP_ISINTEGRAL_HPP
//------------------------------------------------------------------------------
namespace FastCpp
{
//
// BLOCK_SIZE   == 2
// UPBLOCK_SIZE == 4
//
// Capacity is BLOCK_SIZE * UPBLOCK_SIZE * UPBLOCK_SIZE == 2 * 4 * 4 == 32
//
//
//                  +-+-+   +-+-+   +-+-+   +-+-+    +-+-+   +-+-+   +-+-+   +-+-+    +-+-+   +-+-+   +-+-+   +-+-+    +-+-+   +-+-+   +-+-+   +-+-+
// BTreeArrayBlock  | | |   | | |   | | |   | | |    | | |   | | |   | | |   | | |    | | |   | | |   | | |   | | |    | | |   | | |   | | |   | | |
//                  +-+-+   +-+-+   +-+-+   +-+-+    +-+-+   +-+-+   +-+-+   +-+-+    +-+-+   +-+-+   +-+-+   +-+-+    +-+-+   +-+-+   +-+-+   +-+-+
//                   \       \      |      /          \       \      |      /          \       \      |      /          \       \      |      /
//                     \      \     |     /             \      \     |     /             \      \     |     /             \      \     |     /
//                       \     \    |    /                \     \    |    /                \     \    |    /                \     \    |    /
//                        \     \   /   /                  \     \   /   /                  \     \   /   /                  \     \   /   /
//                          \    | |   /                     \    | |   /                     \    | |   /                     \    | |   /
//                            \  | |  /                        \  | |  /                        \  | |  /                        \  | |  /
//                            +-+-+-+-+                        +-+-+-+-+                        +-+-+-+-+                        +-+-+-+-+
// BTreeArrayUpBlock          | | | | |                        | | | | |                        | | | | |                        | | | | |
//                            +-+-+-+-+                        +-+-+-+-+                        +-+-+-+-+                        +-+-+-+-+
//                             \                               \                               /                                /
//                               \                              \                             |                               /
//                                 \                             |                            |                             /
//                                   \                           \                           /                            /
//                                     \                          |                         |                           /
//                                       \                        |                         |                         /
//                                         \                       \                       /                        /
//                                           \                      |                     |                       /
//                                             \                    |                     |                     /
//                                               \                   \                   /                    /
//                                                 \                  |                 |                   /
//                                                   \                |                 |                 /
//                                                     \               \               /                /
//                                                       \              |             |               /
//                                                         \            |             |             /
//                                                           \           \           /            / 
//                                                             \          |         |           /
//                                                               \        |         |         /
//                                                                 \       \       /        /
//                                                                   \      |     |       /
//                                                                     \     \   /      /
//                                                                       \    | |     /
//                                                                         \  | |   /
//                                                                         +-+-+-+-+
// BTreeArrayUpBlock                                                       | | | | |
//                                                                         +-+-+-+-+
//                                                                        /
//                                                                       /
//                                                                      /
//                                                                     /
//                                                 BTreeArray::m_pUpper
//                                                                        
// level            maximal capacity
//   1              2 == BLOCK_SIZE                                                                
//   2              8 == BLOCK_SIZE * UPBLOCK_SIZE
//   3             32 == BLOCK_SIZE * UPBLOCK_SIZE * UPBLOCK_SIZE
//   4            128 == BLOCK_SIZE * UPBLOCK_SIZE * UPBLOCK_SIZE * UPBLOCK_SIZE
//   5            512 == BLOCK_SIZE * UPBLOCK_SIZE * UPBLOCK_SIZE * UPBLOCK_SIZE * UPBLOCK_SIZE
//   5           2048 == BLOCK_SIZE * UPBLOCK_SIZE * UPBLOCK_SIZE * UPBLOCK_SIZE * UPBLOCK_SIZE * UPBLOCK_SIZE
//   6           8192 
//   7          32768 
//   8         131072 
//   9         524288 
//  10        2097152 
//  11        8388608 
//  12       33554432 
//  13      134217728 
//  14      536870912 
//  15     2147483648  ! overflow
//  16     8589934592
//  17    17179869184
//
//
//                                              level - 1
//  maximal capacity = BLOCK_SIZE * UPBLOCK_SIZE


// Notes.
// Note N1
//   The UIndex type can be signed or unsigned value.
//   The UINDEX_LIMIT value should be:
//                  * for signed types: -1 or std::numeric_limits<UIndex>::max()
//                  * for unsigned types: std::numeric_limits<UIndex>::max()
//   The UINDEX_LIMIT is used as specific marker. If some variable of the UIndex 
//   type equels to UINDEX_LIMIT, it means that the variable has undefined value.
//   The standard C++ ISO/IEC 14882:2003 has the following declaration for 
//     namespace std {
//       template<class T> class numeric_limits {
//         public:
//           static T max() throw();
//       }
//     }
//   Therefoe we can not use std::numeric_limits<UIndex>::max() as default 
//   value for the template argument (It is a constant).
//   (But the expresion can be used as default 
//   value for the template argument if you use the ISO standard C++ of 2011).
//
//   The standard C++ 3.9.1/4:
//     Unsigned integers, declared unsigned, shall obey the laws of arithmetic
//              n
//     modulo  2  where n is the number of bits in the value representation of
//     that particular size of integer.
//   and 4.7/2
//     If the destination type is unsigned, the resulting value is the least 
//                                                                n
//     unsigned integer congruent to the source integer (modulo  2  where n is 
//     the number of bits used to represent the unsigned type).
//
//
//   For a positive integer N, two integers A and B are said to be congruent
//   modulo N, written:
//     A --- B (mod N)
//   if their difference A - B is an integer multiple of N. The number N is
//   called the modulus of the congruence.
//   (http://en.wikipedia.org/wiki/Modular_arithmetic)
//
//   Therefore we can used the expression UIndex(-1) to represent the 
//   maximal unsigned value of the type UIndex.
//
//   Example:
//     Let UIndex is 'unsigned int'
//     If sizeof(int) == 4 (and the byte consists of eight bits) then
//     the number of bits is 32.
//     UIndex(-1) == (unsigned int)(-1) == the least unsigned integer congruent
//     to 2 power 32 (i.e. congruent to 4294967296). 
//     4294967295 - (-1) == 4294967296, therefore
//     4294967295 == (unsigned int)(-1)
//     Evidently: 4294967295 == 0xFFFFFFFF == UINT_MAX
//     so UIndex(-1) == (unsigned int)(-1) == UINT_MAX
//
//--------------------------------------------------------------BTreeArray class
//| |       |                 |                 |                            |
template <class T, class  UIndex       = size_t,
                   class  BIndex       = unsigned char,
                   BIndex BLOCK_SIZE   = 8,
                   BIndex UPBLOCK_SIZE = 8,
                   class  Allocator    = std::allocator<T>,
                   UIndex UINDEX_LIMIT = UIndex(-1)// see note N1 above
         >
class BTreeArray : public Detail::BTreeArrayInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
{
  public:
    typedef UIndex UIndexType;
    typedef UIndex SizeType;
  public:
    typedef BTreeArrayIterator     <T, UIndex, BIndex, BLOCK_SIZE, Allocator> IteratorType;
    typedef BTreeArrayConstIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator> ConstIteratorType;
    typedef std::reverse_iterator<IteratorType>                               ReverseIteratorType;
    typedef std::reverse_iterator<ConstIteratorType>                          ConstReverseIteratorType;
  private:
    typedef Detail::BTreeArrayBlockInterface  <T, UIndex, BIndex, BLOCK_SIZE, Allocator> BTreeArrayBlockInterfaceType;
    typedef Detail::BTreeArrayUpBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator> BTreeArrayUpBlockInterfaceType;
    typedef Detail::ElementReference          <T, UIndex, BIndex, BLOCK_SIZE, Allocator> ElementReferenceType;
    typedef Detail::BTreeArrayUpBlock<
                                      T,
                                      UIndex,
                                      BIndex,
                                      BLOCK_SIZE,
                                      UPBLOCK_SIZE,
                                      UINDEX_LIMIT,
                                      Allocator
                                     > BTreeArrayUpBlockType;
  public:
    inline                    BTreeArray        ();
    inline                    BTreeArray        (const Allocator&  rAllocator);
    inline                    BTreeArray        (const BTreeArray& rObj);
    inline                    BTreeArray        (
                                                  const BTreeArray& rObj,
                                                  const Allocator&  rAllocator
                                                );
    explicit                  BTreeArray        (SizeType nSize);
                              BTreeArray        (SizeType nSize, const T& rDefaultElemt);
                              BTreeArray        (SizeType nSize, const T& rDefaultElemt, const Allocator& rAllocator);
    template <class InputIterator>
    inline                    BTreeArray        (
                                                  InputIterator itFirst,
                                                  InputIterator itEnd
                                                );
  public:
    virtual                  ~BTreeArray        ();
  //-------------------------------------------------------------------std names
  public:
    typedef          IteratorType                 iterator;
    typedef          ConstIteratorType            const_iterator;
    typedef          ReverseIteratorType          reverse_iterator;
    typedef          ConstReverseIteratorType     const_reverse_iterator;

    typedef          T                            value_type;
    typedef          value_type&                  reference;
    typedef          const value_type&            const_reference;
    typedef          SizeType                     size_type;
    typedef          Allocator                    allocator_type;
    typedef typename Allocator::difference_type   difference_type;
    typedef typename Allocator::pointer           pointer;
    typedef typename Allocator::const_pointer     const_pointer;
  //--------------------------------------------------------std iterator's names
  public:
    inline  iterator          begin             ()
                              {
                                return Begin();
                              }
    inline  const_iterator    begin             () const
                              {
                                return Begin();
                              }
    inline  iterator          end               ()
                              {
                                return End();
                              }
    inline  const_iterator    end               () const
                              {
                                return End();
                              }
    inline  reverse_iterator  rbegin            ()
                              {
                                return RBegin();
                              }
    inline  const_reverse_iterator
                              rbegin            () const
                              {
                                return RBegin();
                              }
    inline  reverse_iterator  rend              ()
                              {
                                return REnd();
                              }
    inline  const_reverse_iterator
                              rend              () const
                              {
                                return REnd();
                              }
    inline  const_iterator    cbegin            () const
                              {
                                return ConstBegin();
                              }
    inline  const_iterator    cend              () const
                              {
                                return ConstEnd();
                              }
    inline  const_reverse_iterator
                              crbegin           () const
                              {
                                return ConstRBegin();
                              }
    inline  const_reverse_iterator
                              crend             () const
                              {
                                return ConstREnd();
                              }
  //----------------------------------------------------------std method's names
  public:
    inline  bool              empty             () const
                              {
                                return IsEmpty();
                              }
    inline  UIndex            size              () const
                              {
                                return GetCount();
                              }
    inline  size_type         max_size          () const
                              {
                                return GetMaxSize();
                              }
    inline  void              clear             ()
                              {
                                Clear();
                              }
    inline  void              push_back         (const T& rElement)
                              {
                                PushBack(rElement);
                              }
    inline  void              pop_back          ()
                              {
                                RemoveLast();
                              }
    inline  void              push_front        (const T& rElement)
                              {
                                PushFront(rElement);
                              }
    inline  void              pop_front         ()
                              {
                                Remove(0);
                              }
    inline  iterator          insert            (const_iterator position, const T& x)
                              {
                                return position.Insert(x);
                              }
    inline  iterator          insert            (
                                                  const_iterator position,
                                                  size_type      n,
                                                  const T&       x
                                                )
                              {
                                return InsertIt(position, n, x);
                              }
    template <class InputIterator>
    inline  iterator          insert            (
                                                  const_iterator position,
                                                  InputIterator  first,
                                                  InputIterator  last
                                                )
                              {
                                typename IsIntegral<InputIterator>::Type selector;
                                return insert(position, first, last, selector);
                              }
    template <class InputIterator>
    inline  iterator          insert            (
                                                  const_iterator      position,
                                                  InputIterator       first,
                                                  InputIterator       last,
                                                  NotIntegralType_Tag //selector
                                                )
                              {
                                return InsertIts<InputIterator>(position, first, last);
                              }
    template <class SomeIntegralType>
    inline  iterator          insert            (
                                                  const_iterator   position,
                                                  SomeIntegralType nCount,
                                                  SomeIntegralType nValue,
                                                  IntegralType_Tag //selector
                                                )
                              {
                                return InsertIt(position, nCount, nValue);
                              }

    inline  iterator          erase             (const_iterator position)
                              {
                                return position.Erase();
                              }
    inline  iterator          erase             (
                                                  const_iterator first,
                                                  const_iterator last
                                                )
                              {
                                return RemoveIts(first, last);
                              }
    inline  void              swap              (BTreeArray& rArr)
                              {
                                Swap(rArr);
                              }
    inline  void              resize            (size_type sz)
                              {
                                Resize(sz);
                              }
    inline  void              resize            (size_type sz, const T& c)
                              {
                                Resize(sz, c);
                              }
    inline  void              shrink_to_fit     ()
                              {
                                ShrinkToFit();
                              }
    inline  reference         at                (size_type n)
                              {
                                return At(n);
                              }
    inline  const_reference   at                (size_type n) const
                              {
                                return At(n);
                              }
    inline  reference         front             ()
                              {
                                return GetFirst();
                              }
    inline  const_reference   front             () const
                              {
                                return GetFirst();
                              }
    inline  reference         back              ()
                              {
                                return GetLast();
                              }
    inline  const_reference   back              () const
                              {
                                return GetLast();
                              }

  //----------------------------------------------------------------------------
  public:
    inline  BTreeArray&       operator =        (const BTreeArray&     rObj);
    inline  BTreeArray&       operator =        (const std::vector<T>& rVector);
            void              Assign            (const BTreeArray&     rObj);
            void              Assign            (
                                                  SizeType nSize,
                                                  const T& rValue
                                                );
    template <class InputIterator>
            void              AssignIts         (
                                                  InputIterator oBegin,
                                                  InputIterator oEnd
                                                );
    inline  void              Swap              (BTreeArray& rObj);
  public:
            void              PushFront         (const T& rElemt);
            UIndex            PushBack          (const T& rElemt);
            void              Insert            (UIndex Index, T tElemt);
            IteratorType      InsertIt          (ConstIteratorType oIterator, const T& rElemt);
            IteratorType      InsertIt          (ConstIteratorType oIterator, SizeType nCount, const T& rElemt);
    template <class InputIterator>
            IteratorType      InsertIts         (
                                                  ConstIteratorType oIterator,
                                                  InputIterator     oFirst,
                                                  InputIterator     oLast
                                                );

    inline  T&                Front             ();       // no throw
    inline  const T&          Front             () const; // no throw
    inline  T&                GetFirst          ();
    inline  const T&          GetFirst          () const;
    inline  const T&          GetFirst          (const T& rIfEmpty) const;
    inline  T&                GetLast           ();
    inline  const T&          GetLast           () const;
    inline  const T&          GetLast           (const T& rIfEmpty) const;
    inline  T&                At                (UIndex nIndex);
    inline  const T&          At                (UIndex nIndex) const;
    inline  T&                Get               (UIndex nIndex);
    inline  const T&          Get               (UIndex nIndex) const;
    inline  T&                operator []       (UIndex nIndex);
    inline  const T&          operator []       (UIndex nIndex) const;
    inline  const T&          RGet              (UIndex nIndex);
            T                 GetOr             (UIndex nIndex, T tIfWrongIndx);
            T                 Replace           (UIndex nIndex, T tBy);
    inline  UIndex            GetCount          () const;
    inline  bool              IsEmpty           () const;
    inline  UIndex            GetMaxSize        () const;
            T                 Remove            (UIndex Index);
            IteratorType      RemoveIt          (ConstIteratorType oIterator);
            IteratorType      RemoveIts         (
                                                  ConstIteratorType oFrom,
                                                  ConstIteratorType oToExclusive
                                                );
            T                 RemoveLast        ();
    inline  T                 RRemove           (UIndex Index);
            void              Remove            (
                                                  UIndex nFirst,
                                                  UIndex nLastInclusive
                                                );
            void              RemoveRange       (
                                                  UIndex nFirst,
                                                  UIndex nLastInclusive
                                                );
            void              CutTailFromIndex  (UIndex nIndex);
            void              CutTail           (SizeType nCount);
            void              Resize            (SizeType nNewSize);
            void              Resize            (SizeType nNewSize, const T& rValue);
            void              Constate          ();
            void              ShrinkToFit       ();
            void              Clear             ();
    inline  Allocator         GetAllocator      () const;
    virtual Allocator&        GetAllocatorRef   ();
    inline  std::vector<T>    ToVector          () const;

  //-------------------------------------------------------------------iterators
  public:
            IteratorType      Iterator          (UIndex Index);
            IteratorType      Begin             ();
            IteratorType      End               ();
            ConstIteratorType Begin             () const;
            ConstIteratorType End               () const;
    inline  ReverseIteratorType
                              RBegin            ();
    inline  ReverseIteratorType
                              REnd              ();
    inline  ConstReverseIteratorType
                              RBegin            () const;
    inline  ConstReverseIteratorType
                              REnd              () const;

    inline  ConstIteratorType ConstBegin        () const;
    inline  ConstIteratorType ConstEnd          () const;
    inline  ConstReverseIteratorType
                              ConstRBegin       () const;
    inline  ConstReverseIteratorType
                              ConstREnd         () const;
  protected:
            void              PushBack          ();
    template <class SomeIntegralType>
            void              Initialize        (
                                                  SomeIntegralType    nSize,
                                                  SomeIntegralType    rElemt,
                                                  IntegralType_Tag    selector
                                                );
    template <class InputIterator>
            void              Initialize        (
                                                  InputIterator      nSize,
                                                  InputIterator      rElemt,
                                                  NotIntegralType_Tag selector
                                                );
  private:
            void              NewUpper          (BTreeArrayBlockInterfaceType* pNew);
    inline  void              RemoveUpper       ();
    inline  void              DeleteUpper       ();
  //--------------------------------------------------BTreeArrayUpBlockInterface
  private:
    virtual void              IncreaseUCount    ();
    virtual void              IncreaseArrayCount();
    virtual void              DecreaseArrayCount();
    virtual void              DecreaseUCount    ();
    virtual void              DecreaseUCount    (UIndex nCount);
    virtual void              UndefineUCount    ();
    virtual UIndex            GetMyIndex        (BTreeArrayBlockInterfaceType* pBlock);
    virtual ElementReferenceType
                              GetPrev           (BTreeArrayBlockInterfaceType* pBlock);
    virtual ElementReferenceType
                              GetNext           (BTreeArrayBlockInterfaceType* pBlock);
    virtual ElementReferenceType
                              GetNext           (
                                                  BTreeArrayBlockInterfaceType* pBlock,
                                                  ptrdiff_t                     nOffset
                                                );
    virtual BTreeArrayBlockInterfaceType*
                              GetPrevSibling    ( 
                                                  BTreeArrayBlockInterfaceType* pBlock,
                                                  UIndex                        nULevel
                                                );
    virtual BTreeArrayBlockInterfaceType*
                              GetNextSibling    ( 
                                                  BTreeArrayBlockInterfaceType* pBlock,
                                                  UIndex                        nULevel
                                                );
    virtual void              DeleteItem        (BTreeArrayBlockInterfaceType* pBlock);
    virtual void              InsertAfter       (
                                                  BTreeArrayBlockInterfaceType* pAfterThis,
                                                  BTreeArrayBlockInterfaceType* pNewBlock
                                                );
  //---------------------------------------------------------BTreeArrayInterface
  public:
    virtual IteratorType      AddBack           (const T& rElemt);
  private:
            BTreeArrayBlockInterfaceType*
                              m_pUpper;
            UIndex            m_nUCount;
            Allocator         m_oAllocator;
};


//------------------------------------------------- inline function declarations

template <class T,
          class    UIndex_X,
          class    BIndex_X,
          BIndex_X BLOCK_SIZE_X,
          BIndex_X UPBLOCK_SIZE_X,
          class    Allocator_X,
          UIndex_X UINDEX_LIMIT_X,
          class    UIndex_Y,
          class    BIndex_Y,
          BIndex_Y BLOCK_SIZE_Y,
          BIndex_Y UPBLOCK_SIZE_Y,
          class    Allocator_Y,
          UIndex_Y UINDEX_LIMIT_Y
         >
inline bool operator == 
(
  const BTreeArray<T, UIndex_X, BIndex_X, BLOCK_SIZE_X, UPBLOCK_SIZE_X, Allocator_X, UINDEX_LIMIT_X>& x,
  const BTreeArray<T, UIndex_Y, BIndex_Y, BLOCK_SIZE_Y, UPBLOCK_SIZE_Y, Allocator_Y, UINDEX_LIMIT_Y>& y
);


template <class T,
          class    UIndex_X,
          class    BIndex_X,
          BIndex_X BLOCK_SIZE_X,
          BIndex_X UPBLOCK_SIZE_X,
          class    Allocator_X,
          UIndex_X UINDEX_LIMIT_X,
          class    UIndex_Y,
          class    BIndex_Y,
          BIndex_Y BLOCK_SIZE_Y,
          BIndex_Y UPBLOCK_SIZE_Y,
          class    Allocator_Y,
          UIndex_Y UINDEX_LIMIT_Y
         >
inline bool operator != 
(
  const BTreeArray<T, UIndex_X, BIndex_X, BLOCK_SIZE_X, UPBLOCK_SIZE_X, Allocator_X, UINDEX_LIMIT_X>& x,
  const BTreeArray<T, UIndex_Y, BIndex_Y, BLOCK_SIZE_Y, UPBLOCK_SIZE_Y, Allocator_Y, UINDEX_LIMIT_Y>& y
);


template <class T,
          class    UIndex_X,
          class    BIndex_X,
          BIndex_X BLOCK_SIZE_X,
          BIndex_X UPBLOCK_SIZE_X,
          class    Allocator_X,
          UIndex_X UINDEX_LIMIT_X,
          class    UIndex_Y,
          class    BIndex_Y,
          BIndex_Y BLOCK_SIZE_Y,
          BIndex_Y UPBLOCK_SIZE_Y,
          class    Allocator_Y,
          UIndex_Y UINDEX_LIMIT_Y
         >
inline bool operator <
(
  const BTreeArray<T, UIndex_X, BIndex_X, BLOCK_SIZE_X, UPBLOCK_SIZE_X, Allocator_X, UINDEX_LIMIT_X>& x,
  const BTreeArray<T, UIndex_Y, BIndex_Y, BLOCK_SIZE_Y, UPBLOCK_SIZE_Y, Allocator_Y, UINDEX_LIMIT_Y>& y
);


template <class T,
          class    UIndex_X,
          class    BIndex_X,
          BIndex_X BLOCK_SIZE_X,
          BIndex_X UPBLOCK_SIZE_X,
          class    Allocator_X,
          UIndex_X UINDEX_LIMIT_X,
          class    UIndex_Y,
          class    BIndex_Y,
          BIndex_Y BLOCK_SIZE_Y,
          BIndex_Y UPBLOCK_SIZE_Y,
          class    Allocator_Y,
          UIndex_Y UINDEX_LIMIT_Y
         >
inline bool operator <=
(
  const BTreeArray<T, UIndex_X, BIndex_X, BLOCK_SIZE_X, UPBLOCK_SIZE_X, Allocator_X, UINDEX_LIMIT_X>& x,
  const BTreeArray<T, UIndex_Y, BIndex_Y, BLOCK_SIZE_Y, UPBLOCK_SIZE_Y, Allocator_Y, UINDEX_LIMIT_Y>& y
);


template <class T,
          class    UIndex_X,
          class    BIndex_X,
          BIndex_X BLOCK_SIZE_X,
          BIndex_X UPBLOCK_SIZE_X,
          class    Allocator_X,
          UIndex_X UINDEX_LIMIT_X,
          class    UIndex_Y,
          class    BIndex_Y,
          BIndex_Y BLOCK_SIZE_Y,
          BIndex_Y UPBLOCK_SIZE_Y,
          class    Allocator_Y,
          UIndex_Y UINDEX_LIMIT_Y
         >
inline bool operator >
(
  const BTreeArray<T, UIndex_X, BIndex_X, BLOCK_SIZE_X, UPBLOCK_SIZE_X, Allocator_X, UINDEX_LIMIT_X>& x,
  const BTreeArray<T, UIndex_Y, BIndex_Y, BLOCK_SIZE_Y, UPBLOCK_SIZE_Y, Allocator_Y, UINDEX_LIMIT_Y>& y
);


template <class T,
          class    UIndex_X,
          class    BIndex_X,
          BIndex_X BLOCK_SIZE_X,
          BIndex_X UPBLOCK_SIZE_X,
          class    Allocator_X,
          UIndex_X UINDEX_LIMIT_X,
          class    UIndex_Y,
          class    BIndex_Y,
          BIndex_Y BLOCK_SIZE_Y,
          BIndex_Y UPBLOCK_SIZE_Y,
          class    Allocator_Y,
          UIndex_Y UINDEX_LIMIT_Y
         >
inline bool operator >=
(
  const BTreeArray<T, UIndex_X, BIndex_X, BLOCK_SIZE_X, UPBLOCK_SIZE_X, Allocator_X, UINDEX_LIMIT_X>& x,
  const BTreeArray<T, UIndex_Y, BIndex_Y, BLOCK_SIZE_Y, UPBLOCK_SIZE_Y, Allocator_Y, UINDEX_LIMIT_Y>& y
);


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline void swap
(
  BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>& x,
  BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>& y
);


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline std::vector<T>& CopyToStdVector
(
  const BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>& oSrc,
  std::vector<T>& rDest
);


//------------------------------------------- public methods of BTreeArray class


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::BTreeArray()
  : m_pUpper (NULL),
    m_nUCount(0   )
{
  assert(1 < BLOCK_SIZE);  // 1 is not supported
  assert(1 < UPBLOCK_SIZE);// 1 does not have a sense
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::BTreeArray(const BTreeArray& rObj)
  : m_pUpper    (NULL             ),
    m_nUCount   (0                ),
    m_oAllocator(rObj.m_oAllocator)
{
  assert(1 < BLOCK_SIZE);  // 1 is not supported
  assert(1 < UPBLOCK_SIZE);// 1 does not have a sense

  if ( NULL != rObj.m_pUpper )
  {
    assert(0 < rObj.m_nUCount);
    m_pUpper  = rObj.m_pUpper->Clone(m_oAllocator);// can throw exception
    m_nUCount = rObj.m_nUCount;
    assert(NULL != m_pUpper);
    m_pUpper->SetUpNode(this);
  }
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::BTreeArray(const Allocator& rAllocator)
  : m_pUpper    (NULL      ),
    m_nUCount   (0         ),
    m_oAllocator(rAllocator)
{
  assert(1 < BLOCK_SIZE);  // 1 is not supported
  assert(1 < UPBLOCK_SIZE);// 1 does not have a sense
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::BTreeArray
(
  const BTreeArray& rObj,
  const Allocator&  rAllocator
)
  : m_pUpper    (NULL      ),
    m_nUCount   (0         ),
    m_oAllocator(rAllocator)
{
  assert(1 < BLOCK_SIZE);  // 1 is not supported
  assert(1 < UPBLOCK_SIZE);// 1 does not have a sense

  if ( NULL != rObj.m_pUpper )
  {
    assert(0 < rObj.m_nUCount);
    m_pUpper  = rObj.m_pUpper->Clone(m_oAllocator);// can throw exception
    m_nUCount = rObj.m_nUCount;
    assert(NULL != m_pUpper);
    m_pUpper->SetUpNode(this);
  }
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::BTreeArray(SizeType nSize)
  : m_pUpper (NULL),
    m_nUCount(0   )
{
  assert(1 < BLOCK_SIZE);  // 1 is not supported
  assert(1 < UPBLOCK_SIZE);// 1 does not have a sense

  if ( 0 < nSize )
  {
    PushBack(T());// add the fist element as usual

    for(UIndex i = 1; i < nSize; ++i)// fast add the other elementes
      PushBack();
  }
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::BTreeArray(SizeType nSize, const T& rDefaultElemt)
  : m_pUpper (NULL),
    m_nUCount(0   )
{
  assert(1 < BLOCK_SIZE);  // 1 is not supported
  assert(1 < UPBLOCK_SIZE);// 1 does not have a sense

  for(UIndex i = 0; i < nSize; ++i)
    PushBack(rDefaultElemt);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::BTreeArray
(
  SizeType         nSize,
  const T&         rDefaultElemt,
  const Allocator& rAllocator
)
  : m_pUpper    (NULL      ),
    m_nUCount   (0         ),
    m_oAllocator(rAllocator)
{
  assert(1 < BLOCK_SIZE);  // 1 is not supported
  assert(1 < UPBLOCK_SIZE);// 1 does not have a sense

  for(UIndex i = 0; i < nSize; ++i)
    PushBack(rDefaultElemt);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
template <class InputIterator>
inline BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::BTreeArray
(
  InputIterator itFirst,
  InputIterator itEnd
)
  : m_pUpper (NULL),
    m_nUCount(0   )
{
  typename IsIntegral<InputIterator>::Type selector;
  Initialize(itFirst, itEnd, selector);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::~BTreeArray()
{
  Clear();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>&
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::operator = (const BTreeArray& rObj)
{
  Assign(rObj);
  return *this;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>&
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::operator = (const std::vector<T>& rVector)
{
  AssignIts(rVector.begin(), rVector.end());
  return *this;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Assign(const BTreeArray& rObj)
{
  if ( this == &rObj )
    return;

  if ( rObj.m_nUCount < m_nUCount )
    CutTail(m_nUCount - rObj.m_nUCount);

  ConstIteratorType itSrc = rObj.Begin();
  for(
      IteratorType it = Begin(), itEnd = End();
      itEnd != it;
      ++it, ++itSrc
     )
  {
    *it = *itSrc;
  }

  for(const ConstIteratorType itSrcEnd = rObj.End(); itSrcEnd != itSrc; ++itSrc)
    PushBack(*itSrc);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Assign
(
  SizeType nSize,
  const T& rValue
)
{
  if ( nSize < GetCount() )
    CutTail(GetCount() - nSize);

  for(IteratorType it = Begin(), itEnd = End(); itEnd != it; ++it)
    *it = rValue;

  for(SizeType i = GetCount(); i < nSize; ++i)
    PushBack(rValue);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
template <class InputIterator>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::AssignIts
(
  InputIterator oBegin,
  InputIterator oEnd
)
{
  IteratorType       it    = Begin();
  const IteratorType itEnd = End();
  for(; itEnd != it && oBegin != oEnd; ++it, ++oBegin)
    *it = *oBegin;

  if ( oBegin != oEnd )
  {
    assert(itEnd == it);
    for(; oBegin != oEnd; ++oBegin)
      PushBack(*oBegin);
  }
  else
    CutTailFromIndex(it.GetIndex());
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline void
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Swap(BTreeArray& rObj)
{
  if ( this == &rObj )
    return;
  
  std::swap(m_pUpper,  rObj.m_pUpper);
  std::swap(m_nUCount, rObj.m_nUCount);
  if ( NULL != rObj.m_pUpper )
    rObj.m_pUpper->SetUpNode(&rObj);

  if ( NULL != m_pUpper )
    m_pUpper->SetUpNode(this);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::PushFront(const T& rElemt)
{
  assert(NULL == m_pUpper ? 0 == m_nUCount : m_nUCount > 0);

  if ( NULL == m_pUpper )
    m_pUpper = new Detail::BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>(m_oAllocator, this);

  m_nUCount++;
  assert(UINDEX_LIMIT != m_nUCount && "overflow?");

  BTreeArrayBlockInterfaceType* pNew = m_pUpper->PushFront(rElemt, m_oAllocator);

  if ( NULL != pNew )
  {
    assert(NULL != m_pUpper);

    BTreeArrayUpBlockType* pUpper = new BTreeArrayUpBlockType(this, m_nUCount);

    pUpper->AddItem(pNew);
    pUpper->AddItem(m_pUpper);
    m_pUpper = pUpper;
  }
  assert(NULL != m_pUpper);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
UIndex BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::PushBack(const T& rElemt)
{
  assert(NULL == m_pUpper ? 0 == m_nUCount : m_nUCount > 0);

  if ( NULL == m_pUpper )
    m_pUpper = new Detail::BTreeArrayBlock<T, UIndex, BIndex, BLOCK_SIZE, Allocator>(m_oAllocator, this);

  m_nUCount++;
  assert(UINDEX_LIMIT != m_nUCount && "overflow?");

  BTreeArrayBlockInterfaceType* pNew = m_pUpper->Add(rElemt, m_oAllocator);

  if ( NULL != pNew )
  {
    assert(NULL != m_pUpper);
    NewUpper(pNew);
  }

  assert(NULL != m_pUpper);
  return m_nUCount;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Insert(UIndex nUIndex, T tElemt)
{
  assert(0 <= nUIndex && nUIndex <= m_nUCount);

  if ( 0 <= nUIndex && nUIndex < m_nUCount )
  {
    assert(NULL != m_pUpper);

    m_nUCount++;
    assert(UINDEX_LIMIT != m_nUCount && "overflow?");

    BTreeArrayBlockInterfaceType* pNew = m_pUpper->Insert(nUIndex, tElemt, NULL, NULL, m_oAllocator);

    if ( NULL != pNew )
      NewUpper(pNew);
  }
  else
    PushBack(tElemt);

  assert(NULL != m_pUpper);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::IteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::InsertIt
(
  ConstIteratorType oIterator, const T& rElemt
)
{
  assert(UINDEX_LIMIT != (1 + m_nUCount) && "overflow?");

  if ( oIterator == ConstEnd() )
  {
    PushBack(rElemt);
    assert(NULL != m_pUpper);
    assert(oIterator == ConstEnd());
    return IteratorType(m_pUpper->Last());
  }

  return oIterator.Insert(rElemt);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::IteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::InsertIt
(
  ConstIteratorType oIterator,
  SizeType          nCount,
  const T&          rElemt
)
{
  if ( nCount <= 0 )
    return Iterator(oIterator.GetIndex());

  if ( oIterator == ConstEnd() )
  {
    assert(0 < nCount);
    IteratorType oFirstInserted = AddBack(rElemt);

    for(UIndex i = 1; i < nCount; ++i)
      PushBack(rElemt);

    return oFirstInserted;
  }

  assert(UINDEX_LIMIT != (1 + m_nUCount) && "overflow?");
  assert(0 < nCount);
  UIndex nFirstInserted = oIterator.Insert(rElemt).GetIndex();

  for(UIndex i = 1; i < nCount; ++i)
  {
    assert(UINDEX_LIMIT != (1 + m_nUCount) && "overflow?");
    oIterator.Insert(rElemt);
  }

  return Iterator(nFirstInserted);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
template <class InputIterator>
typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::IteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::InsertIts
(
  ConstIteratorType oIterator,
  InputIterator     oFirst,
  InputIterator     oLast
)
{
  if ( oFirst == oLast )
    return Iterator(oIterator.GetIndex());

  if ( oIterator == ConstEnd() )
  {
    IteratorType oFirstInserted = AddBack(*oFirst);
    for(++oFirst; oFirst != oLast; ++oFirst)
      PushBack(*oFirst);

    return oFirstInserted;
  }

  assert(UINDEX_LIMIT != (1 + m_nUCount) && "overflow?");

  UIndex nFirstInserted = oIterator.Insert(*oFirst).GetIndex();
  for(++oFirst; oFirst != oLast; ++oFirst)
  {
    assert(UINDEX_LIMIT != (1 + m_nUCount) && "overflow?");
    oIterator.Insert(*oFirst);
  }

  return Iterator(nFirstInserted);
}


// no throw
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Front()
{
  assert(NULL != m_pUpper && "The array is empty");
  return m_pUpper->GetFirst();
}      


// no throw
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline const T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Front() const
{
  assert(NULL != m_pUpper && "The array is empty");
  return const_cast<BTreeArray*>(this)->GetFirst();
}      


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetFirst()
{
  if ( NULL == m_pUpper )
  {
    //assert(false);
    throw std::range_error(std::string("BTreeArray::GetFirst() fail: the array is empty."));
  }

  return m_pUpper->GetFirst();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline const T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetFirst() const
{
  return const_cast<BTreeArray*>(this)->GetFirst();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline const T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetFirst(const T& rIfEmpty) const
{
  if ( NULL == m_pUpper )
    return rIfEmpty;

  return m_pUpper->GetFirst();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetLast()
{
  if ( NULL == m_pUpper )
  {
    //assert(false);
    throw std::range_error(std::string("BTreeArray::GetLast() fail: the array is empty."));
  }

  return m_pUpper->GetLast();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline const T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetLast() const
{
  return const_cast<BTreeArray*>(this)->GetLast();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline const T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetLast(const T& rIfEmpty) const
{
  if ( NULL == m_pUpper )
    return rIfEmpty;

  return m_pUpper->GetLast();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::At(UIndex nUIndex)
{
  if ( m_nUCount <= nUIndex || nUIndex < 0  )
  {
    //assert(0 <= nUIndex && nUIndex < m_nUCount);
    throw std::range_error(std::string("BTreeArray::At() fail: index out of range."));
  }

  assert(NULL != m_pUpper);
  return m_pUpper->Get(nUIndex);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline const T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::At(UIndex nUIndex) const
{
  return const_cast<BTreeArray*>(this)->Get(nUIndex);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Get(UIndex nUIndex)
{
  assert(0 <= nUIndex && nUIndex < m_nUCount);
  assert(NULL != m_pUpper);
  return m_pUpper->Get(nUIndex);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline const T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Get(UIndex nUIndex) const
{
  return const_cast<BTreeArray*>(this)->Get(nUIndex);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::operator [] (UIndex nUIndex)
{
  return Get(nUIndex);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline const T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::operator [] (UIndex nUIndex) const
{
  return const_cast<BTreeArray*>(this)->Get(nUIndex);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline const T& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::RGet(UIndex nUIndex)
{
  assert(0 <= nUIndex && nUIndex < m_nUCount);
  if ( m_nUCount <= nUIndex || nUIndex < 0  )
    throw std::range_error(std::string("BTreeArray::RGet() fail: index out of range."));

  assert(NULL != m_pUpper);
  return m_pUpper->Get(m_nUCount - nUIndex - 1);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
T BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetOr(UIndex nUIndex, T tIfWrongIndx)
{
  if ( 0 <= nUIndex && nUIndex < m_nUCount )
  {
    assert(NULL != m_pUpper);
    return m_pUpper->Get(nUIndex);
  }

  return tIfWrongIndx;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline UIndex BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetCount() const
{
  return m_nUCount;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline bool BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::IsEmpty() const
{
  return 0 == m_nUCount;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline UIndex BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetMaxSize() const
{
  // the call (std::numeric_limits<UIndex>::max)() in the parenthesis to avoid conflict with max macros. See NOMINMAX.
  return 0 < UINDEX_LIMIT ? UINDEX_LIMIT : (std::numeric_limits<UIndex>::max)();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
T BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Replace(UIndex nUIndex, T tBy)
{
  assert(0 <= nUIndex && nUIndex < m_nUCount);

  if ( nUIndex < 0 || nUIndex >= m_nUCount )
  {
    assert(false);
    return tBy;
  }

  assert(NULL != m_pUpper);

  return m_pUpper->Replace(nUIndex, tBy);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
T BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Remove(UIndex nUIndex)
{
  assert(0 <= nUIndex && nUIndex < m_nUCount);

  if ( nUIndex < 0 || m_nUCount <= nUIndex )
    throw std::range_error(std::string("BTreeArray::Remove() fail: index out of range."));

  assert(NULL != m_pUpper);

  T tRet;
  bool bSubBlockPacked;
  m_nUCount--;
  //UIndex nULevel = 
  m_pUpper->Delete(nUIndex, tRet, bSubBlockPacked, m_oAllocator);

  assert(0 == m_nUCount ? 0 == m_pUpper->GetItemsCount()
                        : 0 != m_pUpper->GetItemsCount()
        );
  assert(0 == m_pUpper->GetItemsCount() ? 0 == m_nUCount : 0 != m_nUCount);

  if ( 0 >= m_nUCount )
    DeleteUpper();
  else
  {
    if ( bSubBlockPacked && 1 == m_pUpper->GetItemsCount() )
      RemoveUpper();
  }

  assert(0 == m_nUCount ? NULL == m_pUpper : NULL != m_pUpper);
  return tRet;
}


// Removes the element that is specified by oIterator from the array.
// Returns an iterator that designates the first element remaining beyond of
//         removed element, or a iterator points to the end of the array if 
//         no such element exists. 
//
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::IteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::RemoveIt
(
  ConstIteratorType oIterator
)
{
  return oIterator.Erase();
}


// Removes the [oFrom, oToExclusive) range of elements from the array.
//   oFrom        - position of the first element removed from the array. 
//   oToExclusive - position just beyond the last element removed from the array.
//
// Returns an iterator that designates the first element remaining beyond any
//         elements removed, or a iterator points to the end of the array if 
//         no such element exists.
//
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::IteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::RemoveIts
(
  ConstIteratorType oFrom,
  ConstIteratorType oToExclusive
)
{
  UIndex nFirst = oFrom.GetIndex();
  UIndex nLast  = oToExclusive.GetIndex();

  if ( nLast <= nFirst )
    return Iterator(nFirst);

  assert(nFirst + 1 <= nLast);
  RemoveRange(nFirst, nLast - 1);
  return Iterator(nFirst);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
T BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::RemoveLast()
{
  if ( 0 == m_nUCount )
    throw std::range_error(std::string("BTreeArray::RemoveLast() fail: array is empty"));

  assert(0 != m_pUpper);

  m_nUCount--;
  T tRet;
  m_pUpper->RemoveLast(tRet, m_oAllocator);

  assert(0 == m_nUCount ? 0 == m_pUpper->GetItemsCount()
                        : 0 != m_pUpper->GetItemsCount()
        );
  assert(0 == m_pUpper->GetItemsCount() ? 0 == m_nUCount : 0 != m_nUCount);

  if ( 0 >= m_nUCount )
    DeleteUpper();

  assert(0 == m_nUCount ? NULL == m_pUpper : NULL != m_pUpper);
  return tRet;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline T BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::RRemove(UIndex nIndex)
{
  assert(0 <= nIndex && nIndex < m_nUCount);

  return Remove(m_nUCount - 1 - nIndex);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Remove(UIndex nFirst, UIndex nLast)
{
  if ( m_nUCount <= 0 )
    throw std::range_error(std::string("BTreeArray::Remove(nFirst, nLast) fail: array is empty"));

  if ( nFirst < 0 )
    throw std::range_error(std::string("BTreeArray::Remove(nFirst, nLast) fail: nFirst is negative"));

  if ( m_nUCount <= nLast )
    throw std::range_error(std::string("BTreeArray::Remove(nFirst, nLast) fail: nLast is out of range"));

  if ( nLast < nFirst )
    return;

  RemoveRange(nFirst, nLast);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::RemoveRange(UIndex nFirst, UIndex nLast)
{
  assert(0 < m_nUCount);
  assert(0 <= nFirst);
  assert(nFirst <= nLast);
  assert(nLast < m_nUCount);

  m_nUCount -= nLast - nFirst + 1;
  m_pUpper->DeleteRange(nFirst, nLast, m_oAllocator);
  if ( 0 == m_pUpper->GetItemsCount() )
    DeleteUpper();
  else
    RemoveUpper();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::CutTailFromIndex(UIndex nUIndex)
{
  if ( nUIndex <= 0 )
    Clear();
  else
  {
    if ( nUIndex < m_nUCount )
    {
      assert(NULL != m_pUpper);
      //const UIndex nULevel = 
      m_pUpper->CutTailFromIndex(nUIndex, m_oAllocator);
      m_nUCount = nUIndex;
      RemoveUpper();
    }
  }
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::CutTail(SizeType nCount)
{
  if ( GetCount() <= nCount )
    Clear();
  else
  {
    if ( 0 < nCount )
      CutTailFromIndex(GetCount() - nCount);
  }
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Resize(SizeType nNewSize)
{
  if ( nNewSize <= m_nUCount )
    CutTail(m_nUCount - nNewSize);
  else
  {
    for(UIndex i = m_nUCount; i < nNewSize; ++i)// fill the tail
      PushBack();
  }
  assert(nNewSize == m_nUCount);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Resize(SizeType nNewSize, const T& rValue)
{
  if ( nNewSize <= m_nUCount )
    CutTail(m_nUCount - nNewSize);
  else
  {
    for(UIndex i = m_nUCount; i < nNewSize; ++i)// fill the tail
      PushBack(rValue);
  }
  assert(nNewSize == m_nUCount);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Constate()
{
  if ( NULL != m_pUpper )
    m_pUpper->Constate();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ShrinkToFit()
{
  if ( NULL == m_pUpper )
  {
    assert(0 == m_nUCount);
    return;
  }

  m_pUpper->ShrinkToFit(m_oAllocator);
  RemoveUpper();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Clear()
{
  assert(0 == m_nUCount ? NULL == m_pUpper : NULL != m_pUpper);

  if ( NULL != m_pUpper )
  {
    m_pUpper->DestroyAndDeallocate(m_oAllocator);
    delete m_pUpper;
    m_pUpper = NULL;
  }
  m_nUCount = 0;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline Allocator BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetAllocator() const
{
  return m_oAllocator;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline Allocator& BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetAllocatorRef()
{
  return m_oAllocator;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline std::vector<T> BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ToVector() const
{
  std::vector<T> oVector;
  return CopyToStdVector(*this, oVector);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::IteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Iterator(UIndex nUIndex)
{
  assert(0 <= nUIndex && nUIndex <= m_nUCount);
  if ( m_nUCount < nUIndex || nUIndex < 0  )
    throw std::range_error(std::string("BTreeArray::Iterator() fail: index out of range."));

  if ( m_nUCount == nUIndex )
    return End();

  return IteratorType(m_pUpper->GetElementReference(nUIndex));
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::IteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Begin()
{
  return IteratorType(NULL != m_pUpper ? m_pUpper->Begin() : ElementReferenceType(this, BLOCK_SIZE));
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::IteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::End()
{
  return IteratorType(ElementReferenceType(this, BLOCK_SIZE));
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ConstIteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Begin() const
{
  return ConstIteratorType(NULL != m_pUpper ? m_pUpper->Begin() : ElementReferenceType(const_cast<BTreeArray*>(this), BLOCK_SIZE));
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ConstIteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::End() const
{
  return ConstIteratorType(ElementReferenceType(const_cast<BTreeArray*>(this), BLOCK_SIZE));
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ReverseIteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::RBegin()
{
  return ReverseIteratorType(End());
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ReverseIteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::REnd()
{
  return ReverseIteratorType(Begin());
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ConstReverseIteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::RBegin() const
{
  return ConstReverseIteratorType(End());
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ConstReverseIteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::REnd() const
{
  return ConstReverseIteratorType(Begin());
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ConstIteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ConstBegin() const
{
  return Begin();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ConstIteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ConstEnd() const
{
  return End();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ConstReverseIteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ConstRBegin() const
{
  return RBegin();
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ConstReverseIteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::ConstREnd() const
{
  return REnd();
}


//---------------------------------------- protected methods of BTreeArray class


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::PushBack()
{
  assert(NULL != m_pUpper);
  assert(UINDEX_LIMIT != (1 + m_nUCount) && "overflow?");

  m_nUCount++;
  BTreeArrayBlockInterfaceType* pNew = m_pUpper->PlusOne(m_oAllocator);

  if ( NULL != pNew )
  {
    assert(NULL != m_pUpper);
    NewUpper(pNew);
  }

  assert(NULL != m_pUpper);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
template <class SomeIntegralType>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Initialize
(
  SomeIntegralType nSize,
  SomeIntegralType rElemt,
  IntegralType_Tag //selector
)
{
  assert(NULL == m_pUpper);
  assert(0 == m_nUCount);

  for(SomeIntegralType i = 0; i < nSize; ++i)
    PushBack(rElemt);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
template <class InputIterator>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::Initialize
(
  InputIterator       itFirst,
  InputIterator       itEnd,
  NotIntegralType_Tag //selector
)
{
  assert(NULL == m_pUpper);
  assert(0 == m_nUCount);

  for(; itFirst != itEnd; ++itFirst)
    PushBack(*itFirst);
}


//------------------------------------------ private methods of BTreeArray class


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::NewUpper
(
  Detail::BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>* pNew
)
{
  assert(NULL != pNew);
  assert(NULL != m_pUpper);

  BTreeArrayUpBlockType* pUpper = new BTreeArrayUpBlockType(this, m_nUCount);

  pUpper->AddItem(m_pUpper);
  pUpper->AddItem(pNew);
  m_pUpper = pUpper;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::RemoveUpper()
{
  assert(NULL != m_pUpper);
  m_pUpper = m_pUpper->RemoveUpper();
  m_pUpper->SetUpNode(this);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::DeleteUpper()
{
  assert(0 == m_nUCount);
  assert(NULL != m_pUpper);

  m_pUpper->DestroyAndDeallocate(m_oAllocator);
  delete m_pUpper;
  m_pUpper = NULL;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::IncreaseUCount()
{
  // nothing to do
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::IncreaseArrayCount()
{
  assert(0 < m_nUCount);
  assert(NULL != m_pUpper);
  assert(UINDEX_LIMIT != (1 + m_nUCount) && "overflow?");
  ++m_nUCount;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::DecreaseArrayCount()
{
  assert(0 < m_nUCount);
  assert(NULL != m_pUpper);
  --m_nUCount;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::DecreaseUCount()
{
  // nothing to do
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::DecreaseUCount(UIndex)
{
  // nothing to do
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::UndefineUCount()
{
  // nothing to do
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
UIndex BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetMyIndex(BTreeArrayBlockInterfaceType* pBlock)
{
  assert(NULL == pBlock || pBlock == m_pUpper);

  return NULL == pBlock ? GetCount() : 0;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
Detail::ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetPrev
(
  Detail::BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>* pItem
)
{
  // Out of range: elemet array[-1] does not exist
  assert(pItem == m_pUpper);
  return Detail::ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>();
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
Detail::ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetNext
(
  Detail::BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>* pItem
)
{
  assert(pItem == m_pUpper);
  return Detail::ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>(this, BLOCK_SIZE);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
Detail::ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetNext
(
  Detail::BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>* pItem, ptrdiff_t nOffset
)
{
  if ( NULL == pItem )
  {
    // call from past-the-end iterator
    assert(nOffset <= 0 && "Out of range");

    if ( 0 == nOffset )
      return Detail::ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>(this, BLOCK_SIZE);

    assert(0 <= GetCount() + nOffset && "Out of range");
    assert(NULL != m_pUpper);

    return m_pUpper->GetElementReference(static_cast<UIndex>(GetCount() + nOffset));
  }

  assert(0 == nOffset && "Out of range");
  assert(pItem == m_pUpper);
  return Detail::ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator>(this, BLOCK_SIZE);
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
Detail::BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetPrevSibling
(
  BTreeArrayBlockInterfaceType* pBlock, UIndex
)
{
  assert(NULL != pBlock);
  //assert(m_pUpper == pBlock);
  return NULL;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
Detail::BTreeArrayBlockInterface<T, UIndex, BIndex, BLOCK_SIZE, Allocator>*
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::GetNextSibling
(
  BTreeArrayBlockInterfaceType* pBlock, UIndex
)
{
  assert(NULL != pBlock);
  //assert(m_pUpper == pBlock);
  return NULL;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::DeleteItem
(
  BTreeArrayBlockInterfaceType* pBlock
)
{
  assert(NULL != pBlock);
  assert(m_pUpper == pBlock);
  assert(m_nUCount <= 1);
  assert(0 == m_pUpper->GetItemsCount());

  delete m_pUpper;
  m_pUpper = 0;
  m_nUCount = 0;
}


//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
void BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::InsertAfter
(
  BTreeArrayBlockInterfaceType* pAfterThis,
  BTreeArrayBlockInterfaceType* pNewBlock
)
{
  assert(NULL != pAfterThis);
  assert(NULL != pNewBlock);
  assert(NULL != m_pUpper);
  assert(pAfterThis == m_pUpper);

  NewUpper(pNewBlock);
}


//-----------------------------------------------------------------------------
//virtual
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
typename BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::IteratorType
BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>::AddBack
(
  const T& rElemt
)
{
  PushBack(rElemt);
  assert(NULL != m_pUpper);
  return m_pUpper->Last();
}


//------------------------------------------------------------- inline functions

template <class T,
          class    UIndex_X,
          class    BIndex_X,
          BIndex_X BLOCK_SIZE_X,
          BIndex_X UPBLOCK_SIZE_X,
          class    Allocator_X,
          UIndex_X UINDEX_LIMIT_X,
          class    UIndex_Y,
          class    BIndex_Y,
          BIndex_Y BLOCK_SIZE_Y,
          BIndex_Y UPBLOCK_SIZE_Y,
          class    Allocator_Y,
          UIndex_Y UINDEX_LIMIT_Y
         >
inline bool operator == 
(
  const BTreeArray<T, UIndex_X, BIndex_X, BLOCK_SIZE_X, UPBLOCK_SIZE_X, Allocator_X, UINDEX_LIMIT_X>& x,
  const BTreeArray<T, UIndex_Y, BIndex_Y, BLOCK_SIZE_Y, UPBLOCK_SIZE_Y, Allocator_Y, UINDEX_LIMIT_Y>& y
)
{
  if ( x.GetCount() != y.GetCount() )
    return false;

#ifdef _MSC_VER
// suppress C4996: 'std::equal': Function call with parameters that may be
// unsafe - this call relies on the caller to check that the passed values
// are correct.
#pragma warning( push )
#pragma warning( disable : 4996 )
#endif//_MSC_VER

  return std::equal(x.Begin(), x.End(), y.Begin());

#ifdef _MSC_VER
#pragma warning( pop ) 
#endif//_MSC_VER
}


template <class T,
          class    UIndex_X,
          class    BIndex_X,
          BIndex_X BLOCK_SIZE_X,
          BIndex_X UPBLOCK_SIZE_X,
          class    Allocator_X,
          UIndex_X UINDEX_LIMIT_X,
          class    UIndex_Y,
          class    BIndex_Y,
          BIndex_Y BLOCK_SIZE_Y,
          BIndex_Y UPBLOCK_SIZE_Y,
          class    Allocator_Y,
          UIndex_Y UINDEX_LIMIT_Y
         >
inline bool operator != 
(
  const BTreeArray<T, UIndex_X, BIndex_X, BLOCK_SIZE_X, UPBLOCK_SIZE_X, Allocator_X, UINDEX_LIMIT_X>& x,
  const BTreeArray<T, UIndex_Y, BIndex_Y, BLOCK_SIZE_Y, UPBLOCK_SIZE_Y, Allocator_Y, UINDEX_LIMIT_Y>& y
)
{
  return !(x == y);
}


template <class T,
          class    UIndex_X,
          class    BIndex_X,
          BIndex_X BLOCK_SIZE_X,
          BIndex_X UPBLOCK_SIZE_X,
          class    Allocator_X,
          UIndex_X UINDEX_LIMIT_X,
          class    UIndex_Y,
          class    BIndex_Y,
          BIndex_Y BLOCK_SIZE_Y,
          BIndex_Y UPBLOCK_SIZE_Y,
          class    Allocator_Y,
          UIndex_Y UINDEX_LIMIT_Y
         >
inline bool operator <
(
  const BTreeArray<T, UIndex_X, BIndex_X, BLOCK_SIZE_X, UPBLOCK_SIZE_X, Allocator_X, UINDEX_LIMIT_X>& x,
  const BTreeArray<T, UIndex_Y, BIndex_Y, BLOCK_SIZE_Y, UPBLOCK_SIZE_Y, Allocator_Y, UINDEX_LIMIT_Y>& y
)
{
  return std::lexicographical_compare(x.Begin(), x.End(), y.Begin(), y.End());
}


template <class T,
          class    UIndex_X,
          class    BIndex_X,
          BIndex_X BLOCK_SIZE_X,
          BIndex_X UPBLOCK_SIZE_X,
          class    Allocator_X,
          UIndex_X UINDEX_LIMIT_X,
          class    UIndex_Y,
          class    BIndex_Y,
          BIndex_Y BLOCK_SIZE_Y,
          BIndex_Y UPBLOCK_SIZE_Y,
          class    Allocator_Y,
          UIndex_Y UINDEX_LIMIT_Y
         >
inline bool operator <=
(
  const BTreeArray<T, UIndex_X, BIndex_X, BLOCK_SIZE_X, UPBLOCK_SIZE_X, Allocator_X, UINDEX_LIMIT_X>& x,
  const BTreeArray<T, UIndex_Y, BIndex_Y, BLOCK_SIZE_Y, UPBLOCK_SIZE_Y, Allocator_Y, UINDEX_LIMIT_Y>& y
)
{
  // (x <= y)  <=>  !(x > y)  <=>  !(y < x)
  return !(y < x);
}


template <class T,
          class    UIndex_X,
          class    BIndex_X,
          BIndex_X BLOCK_SIZE_X,
          BIndex_X UPBLOCK_SIZE_X,
          class    Allocator_X,
          UIndex_X UINDEX_LIMIT_X,
          class    UIndex_Y,
          class    BIndex_Y,
          BIndex_Y BLOCK_SIZE_Y,
          BIndex_Y UPBLOCK_SIZE_Y,
          class    Allocator_Y,
          UIndex_Y UINDEX_LIMIT_Y
         >
inline bool operator >
(
  const BTreeArray<T, UIndex_X, BIndex_X, BLOCK_SIZE_X, UPBLOCK_SIZE_X, Allocator_X, UINDEX_LIMIT_X>& x,
  const BTreeArray<T, UIndex_Y, BIndex_Y, BLOCK_SIZE_Y, UPBLOCK_SIZE_Y, Allocator_Y, UINDEX_LIMIT_Y>& y
)
{
  // (x > y)  <=>  (y < x)
  return y < x;
}


template <class T,
          class    UIndex_X,
          class    BIndex_X,
          BIndex_X BLOCK_SIZE_X,
          BIndex_X UPBLOCK_SIZE_X,
          class    Allocator_X,
          UIndex_X UINDEX_LIMIT_X,
          class    UIndex_Y,
          class    BIndex_Y,
          BIndex_Y BLOCK_SIZE_Y,
          BIndex_Y UPBLOCK_SIZE_Y,
          class    Allocator_Y,
          UIndex_Y UINDEX_LIMIT_Y
         >
inline bool operator >=
(
  const BTreeArray<T, UIndex_X, BIndex_X, BLOCK_SIZE_X, UPBLOCK_SIZE_X, Allocator_X, UINDEX_LIMIT_X>& x,
  const BTreeArray<T, UIndex_Y, BIndex_Y, BLOCK_SIZE_Y, UPBLOCK_SIZE_Y, Allocator_Y, UINDEX_LIMIT_Y>& y
)
{
  // (x >= y)  <=>  !(x < y) 
  return !(x < y);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline void swap
(
  BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>& x,
  BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>& y
)
{
  x.Swap(y);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, BIndex UPBLOCK_SIZE, class Allocator, UIndex UINDEX_LIMIT>
inline std::vector<T>& CopyToStdVector
(
  const BTreeArray<T, UIndex, BIndex, BLOCK_SIZE, UPBLOCK_SIZE, Allocator, UINDEX_LIMIT>& oSrc,
  std::vector<T>& rDest
)
{
  rDest.resize(oSrc.GetCount());
  std::copy(oSrc.Begin(), oSrc.End(), rDest.begin());
  return rDest;
}


//------------------------------------------------------------------------------
}// namespace FastCpp
#endif//FastCpp_BTREEARRAY_HPP
//** (END OF FILE  : BTreeArray.hpp) *******************************************
