//******************************************************************************
// FILE..........: BTreeArrayIterator.hpp
// DESCRIPTION...: iterator for BTreeArray
// AUTHOR........: Dmitry Vinogradov
// COPYRIGHT.....: Copyright(c) FastCpp 2012-2014, Dmitry Vinogradov
// LICENSE.......: "2 EUR" or "advertising note at the about box"
//                 See http://www.FastCpp.com/license.html
//
// HISTORY.......: DATE       COMMENT
//                 ---------- --------------------------------------------------
//                     y  m  d
//                 2013-02-25 Created by VDima.
//******************************************************************************

//---------------------------------------------------------------- include guard
#ifndef FastCpp_BTREEARRAYITERATOR_HPP
#define FastCpp_BTREEARRAYITERATOR_HPP

//--------------------------------------------------------------- import std lib
#include <iterator>
 
//------------------------------------------------------------------------------
#ifndef                     FastCpp_BTREEARRAYUPBLOCK_HPP
#include <FastCpp/BTreeArray/detail/BTreeArrayUpBlock.hpp>
#endif//                    FastCpp_BTREEARRAYUPBLOCK_HPP
//------------------------------------------------------------------------------

namespace FastCpp
{

//----------------------------------------------------- BTreeArrayIterator class
//| |       |                 |                 |                            |
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
class BTreeArrayIterator : public std::iterator<std::bidirectional_iterator_tag, T, ptrdiff_t, T*, T&>
{
  private:
    typedef Detail::ElementReference<T, UIndex, BIndex, BLOCK_SIZE, Allocator> ElementReferenceType;
  public:
    typedef typename std::iterator<std::bidirectional_iterator_tag, T, ptrdiff_t, T*, T&>::difference_type DifferenceType;
  public:
    inline                    BTreeArrayIterator();
    inline                    BTreeArrayIterator(const ElementReferenceType& rRef);
  public:
    inline  UIndex            GetIndex          () const;
    inline  BTreeArrayIterator
                              Insert            (const T& rElemt);
    inline  BTreeArrayIterator
                              Erase             ();
  public:
    inline  T&                operator *        () const;
    inline  T*                operator ->       () const;
    inline  T&                operator []       (DifferenceType nOffset) const;

    inline  bool              operator ==       (const BTreeArrayIterator& rObj) const;
    inline  bool              operator !=       (const BTreeArrayIterator& rObj) const;
    inline  BTreeArrayIterator&
                              operator ++       ();    // prefix
    inline  BTreeArrayIterator
                              operator ++       (int); // postfix
    inline  BTreeArrayIterator&
                              operator --       ();    // prefix
    inline  BTreeArrayIterator
                              operator --       (int); // postfix
    inline  BTreeArrayIterator&
                              operator +=       (DifferenceType nOffset);
    inline  BTreeArrayIterator
                              operator +        (DifferenceType nOffset) const;
    inline  BTreeArrayIterator&
                              operator -=       (DifferenceType nOffset);
    inline  BTreeArrayIterator
                              operator -        (DifferenceType nOffset) const;
    inline  DifferenceType    operator -        (const BTreeArrayIterator& rIterator) const;

  private:
    inline  bool              Eq                (const BTreeArrayIterator& rObj) const;
  private:
            ElementReferenceType
                              m_oElementReference;
};


//--------------------------------------------------------------- public methods


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::BTreeArrayIterator()
  : m_oElementReference()
{
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::BTreeArrayIterator(const ElementReferenceType& rRef)
  : m_oElementReference(rRef)
{
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline UIndex BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::GetIndex() const
{
  if ( BLOCK_SIZE == m_oElementReference.m_nBIndex )
  {
    // this is past-the-end iterator
    assert(NULL != m_oElementReference.m_uRef.m_pArray && "invalid iterator");
    return m_oElementReference.m_uRef.m_pArray->GetMyIndex(NULL);
  }
  assert(NULL != m_oElementReference.m_uRef.m_pBlock);

  return const_cast<BTreeArrayIterator*>(this)->m_oElementReference.m_uRef.m_pBlock->GetUIndex(m_oElementReference.m_nBIndex);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Insert(const T& rElemt)
{
  if ( BLOCK_SIZE == m_oElementReference.m_nBIndex )
  {
    // this is past-the-end iterator
    assert(NULL != m_oElementReference.m_uRef.m_pArray && "invalid iterator");
    return m_oElementReference.m_uRef.m_pArray->AddBack(rElemt);
  }
  assert(NULL != m_oElementReference.m_uRef.m_pBlock);
  return m_oElementReference.m_uRef.m_pBlock->Insert(rElemt, m_oElementReference);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Erase()
{
  assert(BLOCK_SIZE != m_oElementReference.m_nBIndex && "out of range");
  assert(NULL != m_oElementReference.m_uRef.m_pBlock);
  return m_oElementReference.m_uRef.m_pBlock->Erase(m_oElementReference);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline T& BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator * () const
{
  assert(BLOCK_SIZE != m_oElementReference.m_nBIndex && "The iterator is past-the-end iterator");
  assert(NULL != m_oElementReference.m_uRef.m_pBlock);

  return m_oElementReference.m_uRef.m_pBlock->Get(m_oElementReference.m_nBIndex);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline T* BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator -> () const
{
  assert(BLOCK_SIZE != m_oElementReference.m_nBIndex && "The iterator is past-the-end iterator");
  assert(NULL != m_oElementReference.m_uRef.m_pBlock);

  return &(operator*());
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline T& BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator [] (DifferenceType nOffset) const
{
  ElementReferenceType oElementRef;

  if ( BLOCK_SIZE == m_oElementReference.m_nBIndex )
  {
    // this is past-the-end iterator
    assert(NULL != m_oElementReference.m_uRef.m_pArray && "invalid iterator");
    assert(nOffset < 0 && "Only negative offset for the past-the-end iterator is allowed");
    oElementRef = m_oElementReference.m_uRef.m_pArray->GetNext(NULL, nOffset);
  }
  else
  {
    assert(NULL != m_oElementReference.m_uRef.m_pBlock);
    oElementRef = m_oElementReference;
    m_oElementReference.m_uRef.m_pBlock->Next(oElementRef, nOffset);
  }
  
  return oElementRef.m_uRef.m_pBlock->Get(oElementRef.m_nBIndex);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline bool BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator == (const BTreeArrayIterator& rObj) const
{
  return Eq(rObj);
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline bool BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator != (const BTreeArrayIterator& rObj) const
{
  return !Eq(rObj);
}


// ++ prefix
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>&
BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator ++ ()
{
  assert(BLOCK_SIZE != m_oElementReference.m_nBIndex && "The iterator is past-the-end iterator");
  assert(NULL != m_oElementReference.m_uRef.m_pBlock);

  m_oElementReference.m_uRef.m_pBlock->Next(m_oElementReference);

  return *this;
}


// postfix ++
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator ++ (int) 
{
  assert(BLOCK_SIZE != m_oElementReference.m_nBIndex && "The iterator is past-the-end iterator");
  assert(NULL != m_oElementReference.m_uRef.m_pBlock);

  BTreeArrayIterator oRet(*this);

  m_oElementReference.m_uRef.m_pBlock->Next(m_oElementReference);
  
  return oRet;
}


// -- prefix
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>&
BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator -- ()
{
  if ( BLOCK_SIZE == m_oElementReference.m_nBIndex )
  {
    // this is past-the-end iterator
    assert(NULL != m_oElementReference.m_uRef.m_pArray && "invalid iterator");
    m_oElementReference = m_oElementReference.m_uRef.m_pArray->GetNext(NULL, -1);
  }
  else
  {
    assert(NULL != m_oElementReference.m_uRef.m_pBlock);
    m_oElementReference.m_uRef.m_pBlock->Prev(m_oElementReference);
    // The before-the-front iterator does not exist:
    assert(BLOCK_SIZE != m_oElementReference.m_nBIndex && "Out of range: elemet array[-1] does not exist");
  }
  
  return *this;
}


// postfix
template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator -- (int)
{
  BTreeArrayIterator oRet(*this);
  --*this;
  return oRet;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>&
BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator += (DifferenceType nOffset)
{
  if ( BLOCK_SIZE == m_oElementReference.m_nBIndex )
  {
    // this is past-the-end iterator
    assert(NULL != m_oElementReference.m_uRef.m_pArray && "invalid iterator");
    assert(nOffset <= 0 && "past-the-end iterator can be moved only to head");
    m_oElementReference = m_oElementReference.m_uRef.m_pArray->GetNext(NULL, nOffset);
  }
  else
  {
    assert(NULL != m_oElementReference.m_uRef.m_pBlock);
    m_oElementReference.m_uRef.m_pBlock->Next(m_oElementReference, nOffset);
  }
  
  return *this;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator + (DifferenceType nOffset) const
{
  BTreeArrayIterator oRet(*this);
  return oRet += nOffset;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>&
BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator -= (DifferenceType nOffset)
{
  return *this += -nOffset;
}                              


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>
BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator - (DifferenceType nOffset) const
{
  BTreeArrayIterator oRet(*this);
  return oRet -= nOffset;
}


template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline typename BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::DifferenceType
BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::operator - (const BTreeArrayIterator& rIterator) const
{
  return this->GetIndex() - rIterator.GetIndex();
}


//-------------------------------------------------------------- private methods

template <class T, class UIndex, class BIndex, BIndex BLOCK_SIZE, class Allocator>
inline bool BTreeArrayIterator<T, UIndex, BIndex, BLOCK_SIZE, Allocator>::Eq(const BTreeArrayIterator& rObj) const
{
  return m_oElementReference == rObj.m_oElementReference;
}


}// namespace FastCpp
#endif//FastCpp_BTREEARRAYITERATOR_HPP
//** (END OF FILE  : BTreeArrayIterator.hpp) ***********************************
