//*****************************************************************************
// FILE..........: UIntPow.hpp             
// DESCRIPTION...: static exponentiation metafunction:
//                    m
//                   N  == UIntPow<N, m>::result
//
// AUTHOR........: Dmitry Vinogradov
// COPYRIGHT.....: Copyright(c) FastCpp 2012-2013, Dmitry Vinogradov
// LICENSE.......: "2 EUR" or "advertising note at the about box"
//                 See http://www.FastCpp.com/license.html
//
// HISTORY.......: DATE       COMMENT
//                 ---------- --------------------------------------------------
//                     y  m  d
//                 2013-07-01 Created by Dmitry Vinogradov
//******************************************************************************

//---------------------------------------------------------------- include guard
#ifndef FastCpp_UINTPOW_HPP
#define FastCpp_UINTPOW_HPP

//--------------------------------------------------------------- import std lib
#include <limits.h>
//------------------------------------------------------------------------------
namespace FastCpp
{

//                                    N times
//                          /---------^---------\
// UIntPow<a, N>::result == (a * a * a * ... * a)
//
// Example 
//   UIntPow<2, Power>::result gives the table:
//
// 2 raised to the power of 0   ==  UIntPow<2,  0>::result  ==  1
// 2 raised to the power of 1   ==  UIntPow<2,  1>::result  ==  2
// 2 raised to the power of 2   ==  UIntPow<2,  2>::result  ==  4
// 2 raised to the power of 3   ==  UIntPow<2,  3>::result  ==  8
// 2 raised to the power of 4   ==  UIntPow<2,  4>::result  ==  16
// 2 raised to the power of 5   ==  UIntPow<2,  5>::result  ==  32
// 2 raised to the power of 6   ==  UIntPow<2,  6>::result  ==  64
// 2 raised to the power of 7   ==  UIntPow<2,  7>::result  ==  128
// 2 raised to the power of 8   ==  UIntPow<2,  8>::result  ==  256
// 2 raised to the power of 9   ==  UIntPow<2,  9>::result  ==  512
// 2 raised to the power of 10  ==  UIntPow<2, 10>::result  ==  1024
// 2 raised to the power of 11  ==  UIntPow<2, 11>::result  ==  2048
// 2 raised to the power of 12  ==  UIntPow<2, 12>::result  ==  4096
// 2 raised to the power of 13  ==  UIntPow<2, 13>::result  ==  8192
// 2 raised to the power of 14  ==  UIntPow<2, 14>::result  ==  16384
// 2 raised to the power of 15  ==  UIntPow<2, 15>::result  ==  32768
// 2 raised to the power of 16  ==  UIntPow<2, 16>::result  ==  65536
// 2 raised to the power of 17  ==  UIntPow<2, 17>::result  ==  131072
// 2 raised to the power of 18  ==  UIntPow<2, 18>::result  ==  262144
// 2 raised to the power of 19  ==  UIntPow<2, 19>::result  ==  524288
// 2 raised to the power of 20  ==  UIntPow<2, 20>::result  ==  1048576
// 2 raised to the power of 21  ==  UIntPow<2, 21>::result  ==  2097152
// 2 raised to the power of 22  ==  UIntPow<2, 22>::result  ==  4194304
// 2 raised to the power of 23  ==  UIntPow<2, 23>::result  ==  8388608
// 2 raised to the power of 24  ==  UIntPow<2, 24>::result  ==  16777216
// 2 raised to the power of 25  ==  UIntPow<2, 25>::result  ==  33554432
// 2 raised to the power of 26  ==  UIntPow<2, 26>::result  ==  67108864
// 2 raised to the power of 27  ==  UIntPow<2, 27>::result  ==  134217728
// 2 raised to the power of 28  ==  UIntPow<2, 28>::result  ==  268435456
// 2 raised to the power of 29  ==  UIntPow<2, 29>::result  ==  536870912
// 2 raised to the power of 30  ==  UIntPow<2, 30>::result  ==  1073741824
// 2 raised to the power of 31  ==  UIntPow<2, 31>::result  ==  2147483648
// 2 raised to the power of 32  ==  UIntPow<2, 32>::result  ==  0           <-- overflow here for 32-bit unsigned int
// 2 raised to the power of 33  ==  UIntPow<2, 33>::result  ==  0
// 2 raised to the power of 34  ==  UIntPow<2, 34>::result  ==  0
// 2 raised to the power of 35  ==  UIntPow<2, 35>::result  ==  0
// 2 raised to the power of 36  ==  UIntPow<2, 36>::result  ==  0
// 2 raised to the power of 37  ==  UIntPow<2, 37>::result  ==  0
// ...
// 2 raised to the power of ...== UIntPow<2,...>::result == 0
//
//
// UIntPow<N, M>::result equals "N raised to the power of M"
//                       equals 0 if integral overflow occures
// 
template<unsigned int NValue, unsigned int NPower>
struct UIntPow {
    static const unsigned int result = UINT_MAX / NValue <= UIntPow<NValue, NPower - 1>::result
                                         ? 0
                                         : NValue * UIntPow<NValue, NPower - 1>::result;
};


template<unsigned int NValue>
struct UIntPow<NValue, 1U> {
    static const unsigned int result = NValue;
};


template<unsigned int NValue>
struct UIntPow<NValue, 0U> {
    static const unsigned int result = 1;
};


// UIntPowEx<N, M, D>::result equals "N raised to the power of M"
//                            equals D if integral overflow occures
// 
template<unsigned int NValue, unsigned int NPower, unsigned int NOverflowValue>
struct UIntPowEx {
    static const unsigned int result = UIntPow<NValue, NPower>::result <= 0
                                         ? NOverflowValue
                                         : UIntPow<NValue, NPower>::result;
};


}// namespace FastCpp
#endif//FastCpp_UINTPOW_HPP
//** (END OF FILE  : UIntPow.hpp) **********************************************
